@@ -1,4 +1,5 @@
-import main
+from . import main
 import sys
+import typing
 
 sys.exit(main.main())
