@@ -5,22 +5,23 @@ import sys
 import pprint
 from pkg_resources import resource_stream
 import ruamel.yaml as yaml
-try:
-    from ruamel.yaml import CSafeLoader as SafeLoader
-except ImportError:
-    from ruamel.yaml import SafeLoader  # type: ignore
 import avro.schema
 from . import validate
 import json
 import urlparse
+import os
 AvroSchemaFromJSONData = avro.schema.make_avsc_object
 # AvroSchemaFromJSONData=avro.schema.SchemaFromJSONData
+from avro.schema import Names, SchemaParseException
 from . import ref_resolver
+from .ref_resolver import Loader, DocumentType
 from .flatten import flatten
 import logging
 from .aslist import aslist
 from . import jsonld_context
+from .sourceline import SourceLine, strip_dup_lineno, add_lc_filename, bullets, relname
 from typing import Any, AnyStr, cast, Dict, List, Tuple, TypeVar, Union
+from ruamel.yaml.comments import CommentedSeq, CommentedMap
 
 _logger = logging.getLogger("salad")
 
@@ -48,7 +49,7 @@ salad_files = ('metaschema.yml',
 
 
 def get_metaschema():
-    # type: () -> Tuple[avro.schema.Names, List[Dict[unicode, Any]], ref_resolver.Loader]
+    # type: () -> Tuple[Names, List[Dict[unicode, Any]], Loader]
     loader = ref_resolver.Loader({
         "Any": "https://w3id.org/cwl/salad#Any",
         "ArraySchema": "https://w3id.org/cwl/salad#ArraySchema",
@@ -162,8 +163,8 @@ def get_metaschema():
     loader.cache["https://w3id.org/cwl/salad"] = rs.read()
     rs.close()
 
-    j = yaml.load(loader.cache["https://w3id.org/cwl/salad"],
-            Loader=SafeLoader)
+    j = yaml.round_trip_load(loader.cache["https://w3id.org/cwl/salad"])  # type: ignore
+    add_lc_filename(j, "metaschema.yml")
     j, _ = loader.resolve_all(j, "https://w3id.org/cwl/salad#")
 
     # pprint.pprint(j)
@@ -177,8 +178,14 @@ def get_metaschema():
     return (sch_names, j, loader)
 
 
-def load_schema(schema_ref, cache=None):
-    # type: (Union[unicode, Dict[unicode, Any]], Dict) -> Tuple[ref_resolver.Loader, Union[avro.schema.Names, avro.schema.SchemaParseException], Dict[unicode, Any], ref_resolver.Loader]
+def load_schema(schema_ref,  # type: Union[CommentedMap, CommentedSeq, unicode]
+                cache=None   # type: Dict
+                ):
+    # type: (...) -> Tuple[Loader, Union[Names, SchemaParseException], Dict[unicode, Any], Loader]
+    """Load a schema that can be used to validate documents using load_and_validate.
+
+    return document_loader, avsc_names, schema_metadata, metaschema_loader"""
+
     metaschema_names, metaschema_doc, metaschema_loader = get_metaschema()
     if cache is not None:
         metaschema_loader.cache.update(cache)
@@ -194,7 +201,7 @@ def load_schema(schema_ref, cache=None):
         schema_doc, metactx)
 
     # Create the loader that will be used to load the target document.
-    document_loader = ref_resolver.Loader(schema_ctx, cache=cache)
+    document_loader = Loader(schema_ctx, cache=cache)
 
     # Make the Avro validation that will be used to validate the target
     # document
@@ -202,19 +209,53 @@ def load_schema(schema_ref, cache=None):
 
     return document_loader, avsc_names, schema_metadata, metaschema_loader
 
-def load_and_validate(document_loader, avsc_names, document, strict):
-    # type: (ref_resolver.Loader, avro.schema.Names, Union[Dict[unicode, Any], unicode], bool) -> Tuple[Any, Dict[unicode, Any]]
-    if isinstance(document, dict):
-        data, metadata = document_loader.resolve_all(document, document["id"])
-    else:
-        data, metadata = document_loader.resolve_ref(document)
 
-    validate_doc(avsc_names, data, document_loader, strict)
+def load_and_validate(document_loader,  # type: Loader
+                      avsc_names,       # type: Names
+                      document,         # type: Union[CommentedMap, unicode]
+                      strict            # type: bool
+                      ):
+    # type: (...) -> Tuple[Any, Dict[unicode, Any]]
+    """Load a document and validate it with the provided schema.
+
+    return data, metadata
+    """
+    try:
+        if isinstance(document, CommentedMap):
+            source = document["id"]
+            data, metadata = document_loader.resolve_all(
+                document, document["id"], checklinks=False)
+        else:
+            source = document
+            data, metadata = document_loader.resolve_ref(
+                document, checklinks=False)
+    except validate.ValidationException as v:
+        raise validate.ValidationException(strip_dup_lineno(str(v)))
+
+    validationErrors = u""
+    try:
+        document_loader.validate_links(data, u"")
+    except validate.ValidationException as v:
+        validationErrors = unicode(v) + "\n"
+
+    try:
+        validate_doc(avsc_names, data, document_loader, strict, source=source)
+    except validate.ValidationException as v:
+        validationErrors += unicode(v)
+
+    if validationErrors:
+        raise validate.ValidationException(validationErrors)
+
     return data, metadata
 
 
-def validate_doc(schema_names, doc, loader, strict):
-    # type: (avro.schema.Names, Union[Dict[unicode, Any], List[Dict[unicode, Any]], unicode], ref_resolver.Loader, bool) -> None
+def validate_doc(schema_names,  # type: Names
+                 doc,           # type: Union[Dict[unicode, Any], List[Dict[unicode, Any]], unicode]
+                 loader,        # type: Loader
+                 strict,        # type: bool
+                 source=None
+                 ):
+    # type: (...) -> None
     has_root = False
     for r in schema_names.names.values():
         if ((hasattr(r, 'get_prop') and r.get_prop(u"documentRoot")) or (
@@ -228,8 +269,10 @@ def validate_doc(schema_names, doc, loader, strict):
 
     if isinstance(doc, list):
         validate_doc = doc
-    elif isinstance(doc, dict):
-        validate_doc = [doc]
+    elif isinstance(doc, CommentedMap):
+        validate_doc = CommentedSeq([doc])
+        validate_doc.lc.add_kv_line_col(0, [doc.lc.line, doc.lc.col])
+        validate_doc.lc.filename = doc.lc.filename
     else:
         raise validate.ValidationException("Document must be dict or list")
 
@@ -241,10 +284,12 @@ def validate_doc(schema_names, doc, loader, strict):
 
     anyerrors = []
     for pos, item in enumerate(validate_doc):
+        sl = SourceLine(validate_doc, pos, unicode)
         success = False
         for r in roots:
             success = validate.validate_ex(
-                r, item, loader.identifiers, strict, foreign_properties=loader.foreign_properties, raise_ex=False)
+                r, item, loader.identifiers, strict,
+                foreign_properties=loader.foreign_properties, raise_ex=False)
             if success:
                 break
 
@@ -258,28 +303,33 @@ def validate_doc(schema_names, doc, loader, strict):
 
                 try:
                     validate.validate_ex(
-                        r, item, loader.identifiers, strict, foreign_properties=loader.foreign_properties, raise_ex=True)
+                        r, item, loader.identifiers, strict,
+                        foreign_properties=loader.foreign_properties,
+                        raise_ex=True)
                 except validate.ClassValidationException as e:
-                    errors = [u"Could not validate `%s` because\n%s" % (
-                        name, validate.indent(str(e), nolead=False))]
+                    errors = [sl.makeError(u"tried `%s` but\n%s" % (
+                        name, validate.indent(str(e), nolead=False)))]
                     break
                 except validate.ValidationException as e:
-                    errors.append(u"Could not validate as `%s` because\n%s" % (
-                        name, validate.indent(str(e), nolead=False)))
+                    errors.append(sl.makeError(u"tried `%s` but\n%s" % (
+                        name, validate.indent(str(e), nolead=False))))
 
-            objerr = u"Validation error at position %i" % pos
+            objerr = sl.makeError(u"Invalid")
             for ident in loader.identifiers:
                 if ident in item:
-                    objerr = u"Validation error in object %s" % (item[ident])
+                    objerr = sl.makeError(
+                        u"Object `%s` is not valid because"
+                        % (relname(item[ident])))
                     break
             anyerrors.append(u"%s\n%s" %
-                             (objerr, validate.indent(u"\n".join(errors))))
+                             (objerr, validate.indent(bullets(errors, "- "))))
     if anyerrors:
-        raise validate.ValidationException(u"\n".join(anyerrors))
+        raise validate.ValidationException(
+            strip_dup_lineno(bullets(anyerrors, "* ")))
 
 
 def replace_type(items, spec, loader, found):
-    # type: (Any, Dict[unicode, Any], ref_resolver.Loader, Set[unicode]) -> Any
+    # type: (Any, Dict[unicode, Any], Loader, Set[unicode]) -> Any
     """ Go through and replace types in the 'spec' mapping"""
 
     items = copy.deepcopy(items)
@@ -331,8 +381,13 @@ def avro_name(url):  # type: (AnyStr) -> AnyStr
 
 Avro = TypeVar('Avro', Dict[unicode, Any], List[Any], unicode)
 
-def make_valid_avro(items, alltypes, found, union=False):
-    # type: (Avro, Dict[unicode, Dict[unicode, Any]], Set[unicode], bool) -> Union[Avro, Dict]
+
+def make_valid_avro(items,          # type: Avro
+                    alltypes,       # type: Dict[unicode, Dict[unicode, Any]]
+                    found,          # type: Set[unicode]
+                    union=False     # type: bool
+                    ):
+    # type: (...) -> Union[Avro, Dict]
     items = copy.deepcopy(items)
     if isinstance(items, dict):
         if items.get("name"):
@@ -365,13 +420,13 @@ def make_valid_avro(items, alltypes, found, union=False):
     if union and isinstance(items, (str, unicode)):
         if items in alltypes and avro_name(items) not in found:
             return cast(Dict, make_valid_avro(alltypes[items], alltypes, found,
-                        union=union))
+                                              union=union))
         items = avro_name(items)
     return items
 
 
 def extend_and_specialize(items, loader):
-    # type: (List[Dict[unicode, Any]], ref_resolver.Loader) -> List[Dict[unicode, Any]]
+    # type: (List[Dict[unicode, Any]], Loader) -> List[Dict[unicode, Any]]
     """Apply 'extend' and 'specialize' to fully materialize derived record
     types."""
 
@@ -443,7 +498,8 @@ def extend_and_specialize(items, loader):
 
     for t in n:
         if t.get("abstract") and t["name"] not in extended_by:
-            raise validate.ValidationException("%s is abstract but missing a concrete subtype" % t["name"])
+            raise validate.ValidationException(
+                "%s is abstract but missing a concrete subtype" % t["name"])
 
     for t in n:
         if "fields" in t:
@@ -452,8 +508,10 @@ def extend_and_specialize(items, loader):
     return n
 
 
-def make_avro_schema(i, loader):
-    # type: (List[Dict[unicode, Any]], ref_resolver.Loader) -> Tuple[Union[avro.schema.Names,avro.schema.SchemaParseException], List[Dict[unicode, Any]]]
+def make_avro_schema(i,         # type: List[Dict[unicode, Any]]
+                     loader     # type: Loader
+                     ):
+    # type: (...) -> Tuple[Union[Names, SchemaParseException], List[Dict[unicode, Any]]]
     names = avro.schema.Names()
 
     j = extend_and_specialize(i, loader)
