@@ -1,41 +1,32 @@
+from .util import get_data
 import unittest
 import schema_salad.ref_resolver
 import schema_salad.main
 import schema_salad.schema
 from schema_salad.jsonld_context import makerdf
-from pkg_resources import Requirement, resource_filename, ResolutionError  # type: ignore
 import rdflib
-import ruamel.yaml as yaml
+import ruamel.yaml
 import json
 import os
+from schema_salad.sourceline import cmap
 
 try:
     from ruamel.yaml import CSafeLoader as SafeLoader
 except ImportError:
     from ruamel.yaml import SafeLoader  # type: ignore
 
-
-def get_data(filename):
-    filepath = None
-    try:
-        filepath = resource_filename(
-            Requirement.parse("schema-salad"), filename)
-    except ResolutionError:
-        pass
-    if not filepath or not os.path.isfile(filepath):
-        filepath = os.path.join(os.path.dirname(__file__), os.pardir, filename)
-    return filepath
+from ruamel.yaml.comments import CommentedSeq, CommentedMap
 
 
 class TestSchemas(unittest.TestCase):
     def test_schemas(self):
         l = schema_salad.ref_resolver.Loader({})
 
-        ra, _ = l.resolve_all({
+        ra, _ = l.resolve_all(cmap({
             u"$schemas": ["file://" + get_data("tests/EDAM.owl")],
             u"$namespaces": {u"edam": u"http://edamontology.org/"},
             u"edam:has_format": u"edam:format_1915"
-        }, "")
+        }), "")
 
         self.assertEqual({
             u"$schemas": ["file://" + get_data("tests/EDAM.owl")],
@@ -74,7 +65,7 @@ class TestSchemas(unittest.TestCase):
             argsl=[get_data("tests/Process.yml")]))
 
     def test_jsonld_ctx(self):
-        ldr, _, _, _ = schema_salad.schema.load_schema({
+        ldr, _, _, _ = schema_salad.schema.load_schema(cmap({
             "$base": "Y",
             "name": "X",
             "$namespaces": {
@@ -84,9 +75,9 @@ class TestSchemas(unittest.TestCase):
                 "name": "ExampleType",
                 "type": "enum",
                 "symbols": ["asym", "bsym"]}]
-        })
+        }))
 
-        ra, _ = ldr.resolve_all({"foo:bar": "asym"}, "X")
+        ra, _ = ldr.resolve_all(cmap({"foo:bar": "asym"}), "X")
 
         self.assertEqual(ra, {
             'http://example.com/foo#bar': 'asym'
@@ -106,7 +97,7 @@ class TestSchemas(unittest.TestCase):
             },
             "id": "@id"})
 
-        ra, _ = ldr.resolve_all({
+        ra, _ = ldr.resolve_all(cmap({
             "id": "stuff",
             "inputs": {
                 "zip": 1,
@@ -116,7 +107,7 @@ class TestSchemas(unittest.TestCase):
             "other": {
                 'n': 9
             }
-        }, "http://example2.com/")
+        }), "http://example2.com/")
 
         self.assertEqual("http://example2.com/#stuff", ra["id"])
         for item in ra["inputs"]:
@@ -159,7 +150,7 @@ class TestSchemas(unittest.TestCase):
             },
             "id": "@id"})
 
-        ra, _ = ldr.resolve_all({
+        ra, _ = ldr.resolve_all(cmap({
             "inputs": {
                 "inp": "string",
                 "inp2": "string"
@@ -188,7 +179,7 @@ class TestSchemas(unittest.TestCase):
                     "out": ["out"]
                 }
             }
-        }, "http://example2.com/")
+        }), "http://example2.com/")
 
         self.assertEquals(
             {'inputs': [{
@@ -234,13 +225,15 @@ class TestSchemas(unittest.TestCase):
                 get_data("metaschema/%s_schema.yml" % a))
             with open(get_data("metaschema/%s_src.yml" % a)) as src_fp:
                 src = ldr.resolve_all(
-                    yaml.load(src_fp, Loader=SafeLoader), "", checklinks=False)[0]
+                    ruamel.yaml.round_trip_load(src_fp), "",
+                    checklinks=False)[0]
             with open(get_data("metaschema/%s_proc.yml" % a)) as src_proc:
-                proc = yaml.load(src_proc, Loader=SafeLoader)
+                proc = ruamel.yaml.safe_load(src_proc)
             self.assertEqual(proc, src)
 
     def test_yaml_float_test(self):
-        self.assertEqual(yaml.load("float-test: 2e-10")["float-test"], 2e-10)
+        self.assertEqual(ruamel.yaml.safe_load("float-test: 2e-10")["float-test"],
+                2e-10)
 
     def test_typedsl_ref(self):
         ldr = schema_salad.ref_resolver.Loader({})
@@ -254,16 +247,16 @@ class TestSchemas(unittest.TestCase):
             }
         })
 
-        ra, _ = ldr.resolve_all({"type": "File"}, "")
+        ra, _ = ldr.resolve_all(cmap({"type": "File"}), "")
         self.assertEqual({'type': 'File'}, ra)
 
-        ra, _ = ldr.resolve_all({"type": "File?"}, "")
+        ra, _ = ldr.resolve_all(cmap({"type": "File?"}), "")
         self.assertEqual({'type': ['null', 'File']}, ra)
 
-        ra, _ = ldr.resolve_all({"type": "File[]"}, "")
+        ra, _ = ldr.resolve_all(cmap({"type": "File[]"}), "")
         self.assertEqual({'type': {'items': 'File', 'type': 'array'}}, ra)
 
-        ra, _ = ldr.resolve_all({"type": "File[]?"}, "")
+        ra, _ = ldr.resolve_all(cmap({"type": "File[]?"}), "")
         self.assertEqual(
             {'type': ['null', {'items': 'File', 'type': 'array'}]}, ra)
 
@@ -280,12 +273,12 @@ class TestSchemas(unittest.TestCase):
         }
         ldr.add_context(ctx)
 
-        ra, _ = ldr.resolve_all({
+        ra, _ = ldr.resolve_all(cmap({
             "id": "foo",
             "bar": {
                 "id": "baz"
             }
-        }, "http://example.com")
+        }), "http://example.com")
         self.assertEqual({'id': 'http://example.com/#foo',
                           'bar': {
                               'id': 'http://example.com/#foo/baz'},
@@ -294,12 +287,12 @@ class TestSchemas(unittest.TestCase):
         g = makerdf(None, ra, ctx)
         print(g.serialize(format="n3"))
 
-        ra, _ = ldr.resolve_all({
+        ra, _ = ldr.resolve_all(cmap({
             "location": "foo",
             "bar": {
                 "location": "baz"
             }
-        }, "http://example.com", checklinks=False)
+        }), "http://example.com", checklinks=False)
         self.assertEqual({'location': 'http://example.com/foo',
                           'bar': {
                               'location': 'http://example.com/baz'},
@@ -308,12 +301,12 @@ class TestSchemas(unittest.TestCase):
         g = makerdf(None, ra, ctx)
         print(g.serialize(format="n3"))
 
-        ra, _ = ldr.resolve_all({
+        ra, _ = ldr.resolve_all(cmap({
             "id": "foo",
             "bar": {
                 "location": "baz"
             }
-        }, "http://example.com", checklinks=False)
+        }), "http://example.com", checklinks=False)
         self.assertEqual({'id': 'http://example.com/#foo',
                           'bar': {
                               'location': 'http://example.com/baz'},
@@ -322,12 +315,12 @@ class TestSchemas(unittest.TestCase):
         g = makerdf(None, ra, ctx)
         print(g.serialize(format="n3"))
 
-        ra, _ = ldr.resolve_all({
+        ra, _ = ldr.resolve_all(cmap({
             "location": "foo",
             "bar": {
                 "id": "baz"
             }
-        }, "http://example.com", checklinks=False)
+        }), "http://example.com", checklinks=False)
         self.assertEqual({'location': 'http://example.com/foo',
                           'bar': {
                               'id': 'http://example.com/#baz'},
@@ -337,20 +330,20 @@ class TestSchemas(unittest.TestCase):
         print(g.serialize(format="n3"))
 
     def test_mixin(self):
-        ldr = schema_salad.ref_resolver.Loader({})
-        ra = ldr.resolve_ref({"$mixin": get_data("tests/mixin.yml"), "one": "five"},
-                             base_url="file://" + os.getcwd() + "/tests/")
-        self.assertEqual({'id': 'four', 'one': 'five'}, ra[0])
-
-        ldr = schema_salad.ref_resolver.Loader({"id": "@id"})
         base_url = "file://" + os.getcwd() + "/tests/"
-        ra = ldr.resolve_all([{
+        ldr = schema_salad.ref_resolver.Loader({})
+        ra = ldr.resolve_ref(cmap({"$mixin": get_data("tests/mixin.yml"), "one": "five"}),
+                             base_url=base_url)
+        self.assertEqual({'id': 'four', 'one': 'five'}, ra[0])
+        ldr = schema_salad.ref_resolver.Loader({"id": "@id"})
+
+        ra = ldr.resolve_all(cmap([{
             "id": "a",
             "m": {"$mixin": get_data("tests/mixin.yml")}
         }, {
             "id": "b",
             "m": {"$mixin": get_data("tests/mixin.yml")}
-        }], base_url=base_url)
+        }]), base_url=base_url)
         self.assertEqual([{
             'id': base_url + '#a',
             'm': {
@@ -364,6 +357,19 @@ class TestSchemas(unittest.TestCase):
                 'one': 'two'}
         }], ra[0])
 
+    def test_fragment(self):
+        ldr = schema_salad.ref_resolver.Loader({"id": "@id"})
+        b, _ = ldr.resolve_ref(get_data("tests/frag.yml#foo2"))
+        self.assertEquals({"id": b["id"], "bar":"b2"}, b)
+
+    def test_file_uri(self):
+        # Note: this test probably won't pass on Windows.  Someone with a
+        # windows box should add an alternate test.
+        self.assertEquals("file:///foo/bar%20baz/quux", schema_salad.ref_resolver.file_uri("/foo/bar baz/quux"))
+        self.assertEquals("/foo/bar baz/quux", schema_salad.ref_resolver.uri_file_path("file:///foo/bar%20baz/quux"))
+        self.assertEquals("file:///foo/bar%20baz/quux#zing%20zong", schema_salad.ref_resolver.file_uri("/foo/bar baz/quux#zing zong"))
+        self.assertEquals("/foo/bar baz/quux#zing zong", schema_salad.ref_resolver.uri_file_path("file:///foo/bar%20baz/quux#zing%20zong"))
+
 
 if __name__ == '__main__':
     unittest.main()
