@@ -1,18 +1,31 @@
 import pprint
 import avro.schema
+from avro.schema import Schema
 import sys
 import urlparse
+import re
 from typing import Any, Union
+from .sourceline import SourceLine, lineno_re, bullets, indent
+
 
 class ValidationException(Exception):
     pass
 
+
 class ClassValidationException(ValidationException):
     pass
 
-def validate(expected_schema, datum, identifiers=set(), strict=False, foreign_properties=set()):
-    # type: (avro.schema.Schema, Any, Set[unicode], bool, Set[unicode]) -> bool
-    return validate_ex(expected_schema, datum, identifiers, strict=strict, foreign_properties=foreign_properties, raise_ex=False)
+
+def validate(expected_schema,           # type: Schema
+             datum,                     # type: Any
+             identifiers=set(),         # type: Set[unicode]
+             strict=False,              # type: bool
+             foreign_properties=set()   # type: Set[unicode]
+             ):
+    # type: (...) -> bool
+    return validate_ex(
+        expected_schema, datum, identifiers, strict=strict,
+        foreign_properties=foreign_properties, raise_ex=False)
 
 
 INT_MIN_VALUE = -(1 << 31)
@@ -20,11 +33,6 @@ INT_MAX_VALUE = (1 << 31) - 1
 LONG_MIN_VALUE = -(1 << 63)
 LONG_MAX_VALUE = (1 << 63) - 1
 
-def indent(v, nolead=False):  # type: (Union[str, unicode], bool) -> unicode
-    if nolead:
-        return v.splitlines()[0] + u"\n".join([u"  " + l for l in v.splitlines()[1:]])
-    else:
-        return u"\n".join(["  " + l for l in v.splitlines()])
 
 def friendly(v):  # type: (Any) -> Any
     if isinstance(v, avro.schema.NamedSchema):
@@ -38,11 +46,6 @@ def friendly(v):  # type: (Any) -> Any
     else:
         return v
 
-def multi(v, q=""):  # type: (Union[str, unicode], Union[str, unicode]) -> unicode
-    if '\n' in v:
-        return u"%s%s%s\n" % (q, v, q)
-    else:
-        return u"%s%s%s" % (q, v, q)
 
 def vpformat(datum):  # type: (Any) -> str
     a = pprint.pformat(datum)
@@ -50,9 +53,15 @@ def vpformat(datum):  # type: (Any) -> str
         a = a[0:160] + "[...]"
     return a
 
-def validate_ex(expected_schema, datum, identifiers=None, strict=False,
-                foreign_properties=None, raise_ex=True):
-    # type: (avro.schema.Schema, Any, Set[unicode], bool, Set[unicode], bool) -> bool
+
+def validate_ex(expected_schema,            # type: Schema
+                datum,                      # type: Any
+                identifiers=None,           # type: Set[unicode]
+                strict=False,               # type: bool
+                foreign_properties=None,    # type: Set[unicode]
+                raise_ex=True               # type: bool
+                ):
+    # type: (...) -> bool
     """Determine if a python datum is an instance of a schema."""
 
     if not identifiers:
@@ -68,7 +77,7 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
             return True
         else:
             if raise_ex:
-                raise ValidationException(u"the value `%s` is not null" % vpformat(datum))
+                raise ValidationException(u"the value is not null")
             else:
                 return False
     elif schema_type == 'boolean':
@@ -76,7 +85,7 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
             return True
         else:
             if raise_ex:
-                raise ValidationException(u"the value `%s` is not boolean" % vpformat(datum))
+                raise ValidationException(u"the value is not boolean")
             else:
                 return False
     elif schema_type == 'string':
@@ -87,7 +96,7 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
             return True
         else:
             if raise_ex:
-                raise ValidationException(u"the value `%s` is not string" % vpformat(datum))
+                raise ValidationException(u"the value is not string")
             else:
                 return False
     elif schema_type == 'bytes':
@@ -95,12 +104,13 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
             return True
         else:
             if raise_ex:
-                raise ValidationException(u"the value `%s` is not bytes" % vpformat(datum))
+                raise ValidationException(
+                    u"the value `%s` is not bytes" % vpformat(datum))
             else:
                 return False
     elif schema_type == 'int':
         if ((isinstance(datum, int) or isinstance(datum, long))
-            and INT_MIN_VALUE <= datum <= INT_MAX_VALUE):
+                and INT_MIN_VALUE <= datum <= INT_MAX_VALUE):
             return True
         else:
             if raise_ex:
@@ -109,28 +119,22 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
                 return False
     elif schema_type == 'long':
         if ((isinstance(datum, int) or isinstance(datum, long))
-            and LONG_MIN_VALUE <= datum <= LONG_MAX_VALUE):
+                and LONG_MIN_VALUE <= datum <= LONG_MAX_VALUE):
             return True
         else:
             if raise_ex:
-                raise ValidationException(u"the value `%s` is not long" % vpformat(datum))
+                raise ValidationException(
+                    u"the value `%s` is not long" % vpformat(datum))
             else:
                 return False
     elif schema_type in ['float', 'double']:
         if (isinstance(datum, int) or isinstance(datum, long)
-            or isinstance(datum, float)):
+                or isinstance(datum, float)):
             return True
         else:
             if raise_ex:
-                raise ValidationException(u"the value `%s` is not float or double" % vpformat(datum))
-            else:
-                return False
-    elif isinstance(expected_schema, avro.schema.FixedSchema):
-        if isinstance(datum, str) and len(datum) == expected_schema.size:
-            return True
-        else:
-            if raise_ex:
-                raise ValidationException(u"the value `%s` is not fixed" % vpformat(datum))
+                raise ValidationException(
+                    u"the value `%s` is not float or double" % vpformat(datum))
             else:
                 return False
     elif isinstance(expected_schema, avro.schema.EnumSchema):
@@ -142,28 +146,42 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
                     raise ValidationException(u"'Any' type must be non-null")
                 else:
                     return False
+        if not isinstance(datum, basestring):
+            if raise_ex:
+                raise ValidationException(
+                    u"value is a %s but expected a string" % (type(datum).__name__))
+            else:
+                return False
         if datum in expected_schema.symbols:
             return True
         else:
             if raise_ex:
-                raise ValidationException(u"the value `%s`\n is not a valid symbol in enum %s, expected one of %s" % (vpformat(datum), expected_schema.name, "'" + "', '".join(expected_schema.symbols) + "'"))
+                raise ValidationException(u"the value %s is not a valid %s, expected %s%s" % (vpformat(datum), expected_schema.name,
+                                                                                              "one of " if len(
+                                                                                                  expected_schema.symbols) > 1 else "",
+                                                                                              "'" + "', '".join(expected_schema.symbols) + "'"))
             else:
                 return False
     elif isinstance(expected_schema, avro.schema.ArraySchema):
         if isinstance(datum, list):
             for i, d in enumerate(datum):
                 try:
-                    if not validate_ex(expected_schema.items, d, identifiers, strict=strict, foreign_properties=foreign_properties, raise_ex=raise_ex):
+                    sl = SourceLine(datum, i, ValidationException)
+                    if not validate_ex(expected_schema.items, d, identifiers, strict=strict,
+                                       foreign_properties=foreign_properties,
+                                       raise_ex=raise_ex):
                         return False
                 except ValidationException as v:
                     if raise_ex:
-                        raise ValidationException(u"At position %i\n%s" % (i, indent(str(v))))
+                        raise sl.makeError(
+                            unicode("item is invalid because\n%s" % (indent(str(v)))))
                     else:
                         return False
             return True
         else:
             if raise_ex:
-                raise ValidationException(u"the value `%s` is not a list, expected list of %s" % (vpformat(datum), friendly(expected_schema.items)))
+                raise ValidationException(u"the value is not a list, expected list of %s" % (
+                    friendly(expected_schema.items)))
             else:
                 return False
     elif isinstance(expected_schema, avro.schema.UnionSchema):
@@ -175,34 +193,47 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
             return False
 
         errors = []  # type: List[unicode]
+        checked = []
         for s in expected_schema.schemas:
+            if isinstance(datum, list) and not isinstance(s, avro.schema.ArraySchema):
+                continue
+            elif isinstance(datum, dict) and not isinstance(s, avro.schema.RecordSchema):
+                continue
+            elif isinstance(datum, (bool, int, long, float, basestring)) and isinstance(s, (avro.schema.ArraySchema, avro.schema.RecordSchema)):
+                continue
+            elif datum is not None and s.type == "null":
+                continue
+
+            checked.append(s)
             try:
-                validate_ex(s, datum, identifiers, strict=strict, foreign_properties=foreign_properties, raise_ex=True)
+                validate_ex(s, datum, identifiers, strict=strict,
+                            foreign_properties=foreign_properties, raise_ex=True)
             except ClassValidationException as e:
                 raise
             except ValidationException as e:
                 errors.append(unicode(e))
-
-        raise ValidationException(u"the value %s is not a valid type in the union, expected one of:\n%s" % (
-            multi(vpformat(datum), '`'), u"\n".join([
-                u"- %s, but\n %s" % (
-                    friendly(expected_schema.schemas[i]), indent(multi(errors[i])))
-                for i in range(0, len(expected_schema.schemas))])))
+        if errors:
+            raise ValidationException(bullets(["tried %s but\n%s" % (friendly(
+                checked[i]), indent(errors[i])) for i in range(0, len(errors))], "- "))
+        else:
+            raise ValidationException("value is a %s, expected %s" % (
+                type(datum).__name__, friendly(expected_schema)))
 
     elif isinstance(expected_schema, avro.schema.RecordSchema):
         if not isinstance(datum, dict):
             if raise_ex:
-                raise ValidationException(u"`%s`\n is not a dict" % vpformat(datum))
+                raise ValidationException(u"is not a dict")
             else:
                 return False
 
         classmatch = None
         for f in expected_schema.fields:
-            if f.name == "class":
-                d = datum.get("class")
+            if f.name in ("class",):
+                d = datum.get(f.name)
                 if not d:
                     if raise_ex:
-                        raise ValidationException(u"Missing 'class' field")
+                        raise ValidationException(
+                            u"Missing '%s' field" % (f.name))
                     else:
                         return False
                 if expected_schema.name != d:
@@ -212,7 +243,7 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
 
         errors = []
         for f in expected_schema.fields:
-            if f.name == "class":
+            if f.name in ("class",):
                 continue
 
             if f.name in datum:
@@ -224,13 +255,16 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
                     fieldval = None
 
             try:
-                if not validate_ex(f.type, fieldval, identifiers, strict=strict, foreign_properties=foreign_properties, raise_ex=raise_ex):
+                sl = SourceLine(datum, f.name, unicode)
+                if not validate_ex(f.type, fieldval, identifiers, strict=strict, foreign_properties=foreign_properties,
+                                   raise_ex=raise_ex):
                     return False
             except ValidationException as v:
                 if f.name not in datum:
                     errors.append(u"missing required field `%s`" % f.name)
                 else:
-                    errors.append(u"could not validate field `%s` because\n%s" % (f.name, multi(indent(str(v)))))
+                    errors.append(sl.makeError(u"the `%s` field is not valid because\n%s" % (
+                        f.name, indent(str(v)))))
 
         if strict:
             for d in datum:
@@ -239,21 +273,24 @@ def validate_ex(expected_schema, datum, identifiers=None, strict=False,
                     if d == f.name:
                         found = True
                 if not found:
+                    sl = SourceLine(datum, d, unicode)
                     if d not in identifiers and d not in foreign_properties and d[0] not in ("@", "$"):
                         if not raise_ex:
                             return False
                         split = urlparse.urlsplit(d)
                         if split.scheme:
-                            errors.append(u"could not validate extension field `%s` because it is not recognized and strict is True.  Did you include a $schemas section?" % (d))
+                            errors.append(sl.makeError(
+                                u"unrecognized extension field `%s` and strict is True.  Did you include a $schemas section?" % (d)))
                         else:
-                            errors.append(u"could not validate field `%s` because it is not recognized and strict is True, valid fields are: %s" % (d, ", ".join(fn.name for fn in expected_schema.fields)))
+                            errors.append(sl.makeError(u"invalid field `%s`, expected one of: %s" % (
+                                d, ", ".join("'%s'" % fn.name for fn in expected_schema.fields))))
 
         if errors:
             if raise_ex:
                 if classmatch:
-                    raise ClassValidationException(u"%s record %s" % (classmatch, "\n".join(errors)))
+                    raise ClassValidationException(bullets(errors, "* "))
                 else:
-                    raise ValidationException(u"\n".join(errors))
+                    raise ValidationException(bullets(errors, "* "))
             else:
                 return False
         else:
